#ifndef ContinuousGraph_hpp
#define ContinuousGraph_hpp

#include "AbstractGraph.hpp"
#include <string>
#include <limits>
#include "VertexBuffer.hpp"
#include "precision.hpp"

class ContinuousGraph : public AbstractGraph {
public:
	static size_t splitThreshold;	// the number of values where it should be split up into several vertex buffers
	
	// constructor needed to construct a UnitGraph object
	ContinuousGraph(const fPoint * data, size_t dataSize, QVector4D color = QVector4D(0.0, 0.0, 0.0, 1.0), float lineWidth = 1, QVector3D offset = QVector3D(0.0, 0.0, 0.0), const std::string& description = std::string());

	// for graphs with uniform color
	template<class DataIterator>
	ContinuousGraph(DataIterator dataBegin, DataIterator dataEnd, QVector4D color = QVector4D(0.0, 0.0, 0.0, 1.0), float lineWidth = 1, QVector3D offset = QVector3D(0.0, 0.0, 0.0), const std::string& description = std::string())  :
		AbstractGraph(offset),
		indices_(),
		positions_(),
		description_(description),
		colors_(),
		maxValue_(-std::numeric_limits<fPoint>::infinity()),
		minValue_(std::numeric_limits<fPoint>::infinity()),
		vertexBuffers_(),
		lineWidth_(lineWidth)
	{
		convert(dataBegin, dataEnd, &color, 1);
		calculateExtremeValues();
	}

	// for graphs with varying color
	template<class DataIterator, class ColorIterator>
	ContinuousGraph(DataIterator dataBegin, DataIterator dataEnd, ColorIterator colorsBegin, float lineWidth = 1, QVector3D offset = QVector3D(0.0, 0.0, 0.0),  const std::string& description = std::string()) :
		AbstractGraph(offset),
		indices_(),
		positions_(),
		description_(description),
		colors_(),
		maxValue_(-std::numeric_limits<fPoint>::infinity()),
		minValue_(std::numeric_limits<fPoint>::infinity()),
		vertexBuffers_(),
		lineWidth_(lineWidth)
	{
		convert(dataBegin, dataEnd, colorsBegin, dataEnd - dataBegin);
		calculateExtremeValues();
	}

	~ContinuousGraph();
	
	// change graph color at a certain index on existing data
	template<class ColorIterator>
	void changeColor(ColorIterator colorsBegin, ColorIterator colorsEnd, size_t offset = 0)
	{
		for (size_t i = 0; offset + i < colors_.size() && colorsBegin != colorsEnd; ++i) {
			colors_[offset + i] = *(colorsBegin++);
		}
		
		for (size_t i = 0; i != vertexBuffers_.size(); ++i) {
			if (vertexBuffers_[i]) {
				delete vertexBuffers_[i];
				vertexBuffers_[i] = NULL;		
			}
		}
		vertexBuffers_.clear();
		
		//generateIndicesAndRegenerateVertexBuffer();
	}

	virtual const QVector3D* getDataPtr() const;
	virtual fPoint getMaxValue() const;
	virtual fPoint getMinValue() const;
	virtual std::string getDescription() const;
	virtual size_t getSize() const;

	// generates the index buffer for the given primitive type
	// this method also regenerates the vertexbuffer with the added data
	virtual void generateIndicesAndRegenerateVertexBuffer() = 0;
	
protected:
	// calculates the maximum and minimum value of the data set
	void calculateExtremeValues();
	
	template<class DataIterator, class ColorIterator>
	void convert(DataIterator dataBegin, DataIterator dataEnd, ColorIterator colorsBegin, size_t numColors)
	{
		size_t numValues = dataEnd - dataBegin;
		positions_.resize(numValues);
		colors_.resize(numValues);
		
		for (size_t i = 0; i < numValues; ++i) {
			positions_[i] = QVector3D((float) i, dataBegin[i], 0.0f);
			if (i < numColors) {
				colors_[i] = colorsBegin[i];
			} else {
				colors_[i] = colorsBegin[numColors - 1];
			}
		}
	}
	
	std::vector<int> indices_;
	std::vector<QVector3D> positions_;
	
	std::string description_;
	
	std::vector<QVector4D> colors_;
	
	fPoint maxValue_;
	fPoint minValue_;
	
	std::vector<VertexBuffer*> vertexBuffers_;
	float lineWidth_;
};

#endif
