#if defined(WIN32)
#include <GL/glew.h>
#include <GL/glu.h>
#else
#include <QtOpenGL>
#include <OpenGL/glu.h>
#endif

#include "Grapher.hpp"

#include <sstream>
#include <iostream>
#include <math.h>
#include <limits>

#include <QWheelEvent>
#include <QTimer>
#include <cmath>

#include "pow.hpp"
#include "mathematics.hpp"

Grapher::Grapher(const std::string& xUnit, QWidget *parent) :
QGLWidget(QGLFormat(
					QGL::DoubleBuffer | QGL::DepthBuffer
#if !defined(WIN32)
					| QGL::AlphaChannel
#endif
					| QGL::SampleBuffers)
		  , parent),
	data_(),
	xUnit_(xUnit),
	numValues_(501),
	graphCenter_(0),
	zoomFactor_(1.0f),
	openGLPos_(),
	vecPosition_(),
	dragError_(),
	isFirstFrame_(false),
	pressed_(false),
	leftPressed_(false),
	imageFBO_(NULL),
	valuedist_(0.1f),
	windowOffset_(0),
	hovered_(false),
	hoveredPos_(),
	hoveredGraph_(NULL),
	antialiasing_(false),
	lastNumDivides_(0),
	divideOccurredTimer_(),
	rotated_(false),
	fontSize_(8),
	maxValue_(-std::numeric_limits<fPoint>::infinity()),
	minValue_(std::numeric_limits<fPoint>::infinity()),
	autoScale_(false),
	near_(0.0),
	far_(0.0),
	redrawPending_(true),
	redrawTimer_(),
	autoHover_(true),
	snapSelection_(false),
	viewMatrix_(),
	projectionMatrix_(),
	spacing_(0.0f)
{
	selectionRange_.first = 0;
	selectionRange_.second = 0;
	
	valuedist_ = 1.0f / ((numValues_ - 1) / 2);
	hoveredIndex_ = 0;
	
	this->setMouseTracking(true);
	divideOccurredTimer_.start();
	
	redrawTimer_ = new QTimer(this);
	connect(redrawTimer_, SIGNAL(timeout()), this, SLOT(checkRedrawPending()));
	redrawTimer_->start(30);
	
	setAutoFillBackground(false);
}

Grapher::~Grapher()
{
    clear();
}

void Grapher::resizeGL(int w, int h)
{
    QWidget::resize(w, h);
	
	if (w == 0 || h == 0) {
		return;
	}
	
	projectionMatrix_.setToIdentity();
	projectionMatrix_.ortho(-1.0, 1.0, -1.0, 1.0, near_, far_);
	
    glMatrixMode(GL_PROJECTION);
	glLoadMatrixd(projectionMatrix_.data());
	
    glMatrixMode(GL_MODELVIEW);
	
    glViewport(0, 0, (GLsizei) this->width(), (GLsizei) this->height());
	
	if (imageFBO_) {
		delete imageFBO_;
	}
	
	QGLFramebufferObjectFormat form;
	form.setSamples(16);
	
	imageFBO_ = new QGLFramebufferObject(w, h, form);
}

void Grapher::paintEvent(QPaintEvent* e)
{
    //QGLWidget::paintEvent(e);
	
	makeCurrent();
	paintGL();
	
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	
	// we draw the labels with a separate paint function, so they are properly present in the exported image
	QPainter painter(this);
	QFont widgetFont = font();
	widgetFont.setPointSize(fontSize_);
	painter.setFont(widgetFont);
	drawLabels(painter);
	painter.end();
}

void Grapher::mousePressEvent(QMouseEvent* e)
{
	vecPosition_ = QVector3D(
		(2.0f * float(e->x()) - float(width()))  / float(width()),
		(float(height()) - 2.0f * float(e->y())) / float(height()),
		0.0f
	);
	dragError_ = QVector3D();
	
	setCursor(QCursor(Qt::ClosedHandCursor));
	isFirstFrame_ = true;
	
	if (e->button() == Qt::LeftButton) {
		selectionRange_.first = 0; // unselect selected area
		selectionRange_.second = 0;
		
		leftPressed_ = true;
	} else if (e->button() == Qt::RightButton) {
		pressed_ = true;
	}
}

void Grapher::mouseReleaseEvent(QMouseEvent* e)
{
	isFirstFrame_ = true;
	setCursor(QCursor(Qt::ArrowCursor));
	
	if (e->button() == Qt::LeftButton) {
		leftPressed_ = false;
	} else if (e->button() == Qt::RightButton) {
		pressed_ = false;
		
		// emit selected area as signal (contains index range of selected area of the current displayed window)
		emit currentSelectedIndexes(selectionRange_);
	}
}

void Grapher::mouseDoubleClickEvent(QMouseEvent* e)
{
	QVector3D v(
		(2.0f * float(e->x()) - float(width()))  / float(width()),
		(float(height()) - 2.0f * float(e->y())) / float(height()),
		0.0f
	);
	
	if (rotated_) {
		float tmp = v.x();
		v.setX(v.y());
		v.setY(-tmp);
	}
	
	QVector3D tmp;
	if (rotated_) {
		tmp.setX(-v.y());
		tmp.setY(v.x());
	} else {
		tmp.setX(v.x());
		tmp.setY(v.y());
	}
	
	bool invertible;
	// calculate the point in worldspace under the mouse to be the new grapher center
	QMatrix4x4 viewprojection = (projectionMatrix_ * viewMatrix_).inverted(&invertible);
	QVector3D worldPos = viewprojection * tmp;
	
	center(worldPos.x());
	emit centerChanged(graphCenter_);
}


void Grapher::mouseMoveEvent(QMouseEvent* e)
{
	// v is the position of the mouse
	// cursor over canvas: v ranges in R, from -1 to 1 in both directions, left bottom corner is -1,-1
	// cursor not over canvas: if mouse was pressed before cursor left canvas, position is outside the range R
	QVector3D v(
		(2.0f * float(e->x()) - float(width()))  / float(width()),
		(float(height()) - 2.0f * float(e->y())) / float(height()),
		0.0f
	);

	if (rotated_) {
		float tmp = v.x();
		v.setX(v.y());
		v.setY(-tmp);
	}
	
	if (pressed_ || leftPressed_) {
		if (isFirstFrame_) {
			vecPosition_ = v;
			isFirstFrame_ = false;
			redrawPending_ = true;
			
			// right mouse button, selection: we figure out the first selected value
			if (pressed_) {
				bool invertable;
				QMatrix4x4 viewprojection = (projectionMatrix_ * viewMatrix_).inverted(& invertable);
				
				QVector3D tmp;
				if (rotated_) {
					tmp.setX(-v.y());
					tmp.setY(v.x());
				} else {
					tmp.setX(v.x());
					tmp.setY(v.y());
				}
				
				QVector3D worldPos = viewprojection * tmp;
				
				selectionRange_.first = worldPos.x();
				if (snapSelection_) {
					selectionRange_.first = round(selectionRange_.first);
				}
			}
			
			return;
		}
		
		// some other frame than the other of mouse interaction
		QVector3D vecPosition = v;
		QVector3D vecDifference = vecPosition - vecPosition_;

		if (pressed_) {
			// updated the second value for the selection
			bool invertable;
			QMatrix4x4 viewprojection = (projectionMatrix_ * viewMatrix_).inverted(& invertable);
			
			QVector3D tmp;
			if (rotated_) {
				tmp.setX(-v.y());
				tmp.setY(v.x());
			} else {
				tmp.setX(v.x());
				tmp.setY(v.y());
			}
			
			QVector3D worldPos = viewprojection * tmp;
			selectionRange_.second = worldPos.x();
			
			if (snapSelection_) {
				selectionRange_.second = round(selectionRange_.second);
			}
			// openGLPos_ += vecDifference;
		} else if (leftPressed_) {
			// dragging the graph
			float totaldifference;
			if (vecDifference.x() < 0) {
				totaldifference = vecDifference.x() - dragError_.x();
			} else {
				totaldifference = vecDifference.x() + dragError_.x();
			}
			
			// we drag in integer values, not continuous to get a "snap to value" effect
			if (fabs(totaldifference) > valuedist_) {
				int64_t change = ((int64_t) (-totaldifference / valuedist_));
				int64_t tempCenter = graphCenter_ + change;
				graphCenter_ += change;
				if (tempCenter < 0) {
					graphCenter_ = 0;
				}
				
				double diff = fabs(totaldifference) - fabs(valuedist_ * change);
				
				// keeping track of the drag error when only doing integers in a float world
				dragError_.setX(diff);
				
				vecPosition_ = v;
				
				emit centerChanged(graphCenter_);
			}
		}
	}
	
	if (!leftPressed_) {
		vecPosition_ = v;
	}
	
	if (pressed_ || leftPressed_) {
		hovered_ = false;
		redrawPending_ = true;
		return;
	}
	
	calculateHovered();
	
	redrawPending_ = true;
}

void Grapher::redraw()
{
	this->makeCurrent();
	this->paintGL();
	
	QPainter painter(this);
	painter.beginNativePainting();
	
	drawLabels(painter);
	
	painter.endNativePainting();
	
    this->swapBuffers();
}

UnitGraph* Grapher::addUnitGraph(size_t numValues, fPoint spacing, bool repeat, QVector4D color, float lineWidth, QVector3D offset, const std::string& unit)
{
	UnitGraph * ndata = new UnitGraph(numValues, spacing, repeat, color, lineWidth, offset, unit);
	
	data_.insert(ndata);
	
	spacing_ = spacing;
	
	redrawPending_ = true;
	
	return ndata;
}

void Grapher::removeData(AbstractGraph* graph)
{
	std::set<AbstractGraph*, GraphComparator>::iterator iter = data_.find(graph);
	if (iter == data_.end()) {
		throw std::runtime_error("Attempted to remove a graph that did not exist.");
	}
	data_.erase(graph);
	delete graph;
	
	minValue_ = 0;
	maxValue_ = 0;
	
	std::set<AbstractGraph*, GraphComparator>::iterator i = data_.begin();
	while (i != data_.end()) {
		if ((*i)->getMaxValue() > maxValue_) {
			maxValue_ = (*i)->getMaxValue();
		}
		if ((*i)->getMinValue() < minValue_) {
			minValue_ = (*i)->getMinValue();
		}
		++i;
	}
	redrawPending_ = true;
}

void Grapher::clear()
{
	std::set<AbstractGraph*, GraphComparator>::iterator i = data_.begin();
	while (i != data_.end()) {
		delete *i;
		++i;
	}
	data_.clear();
	minValue_ = 0;
	maxValue_ = 0;
	near_ = 0.0f;
	far_ = 0.0f;
	
	redrawPending_ = true;
}

void Grapher::wheelEvent(QWheelEvent* event)
{
	if (event->modifiers() & Qt::ShiftModifier) {	// vertical zoom
		zoomFactor_ *= (event->delta() > 0) ? 1.25 : 0.8;
		if (zoomFactor_ < 0.1f) {
			zoomFactor_ = 0.1f;
		}
		event->accept();
		redrawPending_ = true;
		return;
	}
	
	setNumValues(numValues_ * ((event->delta() > 0) ? 1.25 : 0.8));
	event->accept();
}

void Grapher::saveTo(QString filePath)
{
	this->makeCurrent();
	imageFBO_->bind();
	paintGL();
	imageFBO_->release();
	
	// labels need to be dealt with separately from normal OpenGL rendering, since we use Qts overpainting for it, thus rendering text like standard Qt
	QImage tempImage = imageFBO_->toImage();
	
	QPainter painter(& tempImage);
	painter.beginNativePainting();
	
	drawLabels(painter);
	
	painter.endNativePainting();
	
	tempImage.save(filePath);
}

int Grapher::getFontSize()
{
	return fontSize_;
}

size_t Grapher::getNumValues() const
{
	return numValues_;
}

size_t Grapher::getCenter() const
{
	return graphCenter_;
}

std::pair<int, int> Grapher::getSelection() const
{
	return selectionRange_;
}

void Grapher::center(size_t frame)
{
    graphCenter_ = frame;
	selectionRange_.first = 0; // unselect selected area
	selectionRange_.second = 0;
    
    redrawPending_ = true;
}

void Grapher::setNumValues(size_t numValues)
{
	if (numValues % 2 == 0) {
		++numValues;
	}
	if (numValues < 3) {
		numValues = 3;
	}
	
	valuedist_ = 1.0f / ((numValues - 1) / 2);
	float factor = 1.0 / numValues_ * numValues;
	
	if (numValues_ != numValues) {	// only emit if there's a change (this is to break signal loops)
		numValues_ = numValues;
		// change selected area according to zoom factor
		emit numValuesChanged(numValues);
	}
	redrawPending_ = true;
}

void Grapher::setAntiAliasing(bool enable)
{
	antialiasing_ = enable;
	
	redrawPending_ = true;
}

void Grapher::setRotated(bool rotated)
{
	rotated_ = rotated;
	
	redrawPending_ = true;
}

void Grapher::setFontSize(int pixel)
{
	fontSize_ = pixel;
	
	redrawPending_ = true;
}

void Grapher::setAutoScale(bool autoScale)
{
	autoScale_ = autoScale;
	
	redrawPending_ = true;
}

void Grapher::setAutoHover(bool autoHover)
{
	autoHover_ = autoHover;
	
	redrawPending_ = true;
}

void Grapher::setSelectedArea(std::pair<float, float> area)
{
	selectionRange_ = area;
	redrawPending_ = true;
}

void Grapher::clearSelection()
{
	setSelectedArea(std::make_pair(0, 0));
}

void Grapher::setSelectionSnap(bool snap)
{
	snapSelection_ = snap;
}

void Grapher::checkRedrawPending()
{
	if (redrawPending_) {
		redrawPending_ = false;
		update();
	}
}

void Grapher::paintGL()
{
	if (autoHover_) {
		calculateHovered();
	}
	
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	
	if (antialiasing_) {
		glEnable(GL_MULTISAMPLE);
	}
	
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	
	// calculating the view and projection matrices based on the information of centering, values displayed and the set data
	viewMatrix_.setToIdentity();
	projectionMatrix_.setToIdentity();
	
	if (!data_.empty()) {
		near_ = -(*data_.begin())->getOffset().z();
		far_ = -(*(--data_.end()))->getOffset().z();
	}
	
	float stepSign = ((far_ - near_) >= 0.0f) ? 1.0f : -1.0f;
	near_ -= 2.0f * stepSign;
	far_ += 2.0f * stepSign;
	
	if (rotated_) {
		viewMatrix_.lookAt(QVector3D((float)graphCenter_, 0.0, 0.0), QVector3D((float)graphCenter_, 0.0, -1.0), QVector3D(1.0, 0.0, 0.0));
		
		if (autoScale_) {
			projectionMatrix_.ortho(-(float)maxValue_ * zoomFactor_, -(float)minValue_ * zoomFactor_, -(float)((numValues_ - 1) / 2), +(float)((numValues_ - 1) / 2), near_, far_);
		} else {
			projectionMatrix_.ortho(-(float)maxValue_ * zoomFactor_, +(float)maxValue_ * zoomFactor_, -(float)((numValues_ - 1) / 2), +(float)((numValues_ - 1) / 2), near_, far_);
		}
	} else {
		viewMatrix_.lookAt(QVector3D((float)graphCenter_, 0.0, 0.0), QVector3D((float)graphCenter_, 0.0, -1.0), QVector3D(0.0, 1.0, 0.0));
		
		if (autoScale_) {
			projectionMatrix_.ortho(-(float)((numValues_ - 1) / 2), +(float)((numValues_ - 1) / 2), (float)minValue_ * zoomFactor_, (float)maxValue_ * zoomFactor_, near_, far_);
		} else {
			projectionMatrix_.ortho(-(float)((numValues_ - 1) / 2), +(float)((numValues_ - 1) / 2), -(float)maxValue_ * zoomFactor_, +(float)maxValue_ * zoomFactor_, near_, far_);
		}
	}
	
	
	glMatrixMode(GL_PROJECTION);
	glLoadMatrixd(projectionMatrix_.data());
	
	glPushMatrix();
	
	
    glMatrixMode(GL_MODELVIEW);
    glLoadMatrixd(viewMatrix_.data());
	
	
	float framex = 0.0;
	size_t midpoint;
	
	int64_t indexoffset = openGLPos_.x() / valuedist_;
	
	// draw the selection
    this->drawSelectedArea();
	
	glPushMatrix();
	
	glDisable(GL_DEPTH_TEST);
	
	// draw every graph
	for (std::set<AbstractGraph*, GraphComparator>::iterator iter = data_.begin(); iter != data_.end(); ++iter) {
		this->drawGraph(*iter, indexoffset, &midpoint);
	}
	
	glEnable(GL_DEPTH_TEST);
	
	glPopMatrix();
	
	float zValue = 0.0f;
	
	fPoint high = maxValue_;
	fPoint low = -maxValue_;
	
	if (autoScale_) {
		low = minValue_ ;
	}
	
	// draw the center line
	glBegin(GL_LINES);
	glColor3f(0.0f, 0.0f, 0.0f);
    glVertex3f(graphCenter_, high, zValue);
    glVertex3f(graphCenter_, low, zValue);
    glEnd();
	glColor3f(0.0f, 0.0f, 0.0f);
	
	// drawing the hovered point, if there is one
	if (hovered_) {
		glPushAttrib(GL_POINT_BIT);
		glPointSize(10.0f);
		glBegin(GL_POINTS);
		glColor3f(0.0f, 0.0f, 1.0f);
		glVertex3f(hoveredTop_.x(), hoveredTop_.y(), zValue);
		glVertex3f(hoveredBottom_.x(), hoveredBottom_.y(), zValue);
		glEnd();
		glPopAttrib();
		
		glBegin(GL_LINES);
		glVertex3f(hoveredTop_.x(), hoveredTop_.y(), zValue);
		glVertex3f(hoveredBottom_.x(), hoveredBottom_.y(), zValue);
		glEnd();
	}
	
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	
	if (antialiasing_) {
		glDisable(GL_MULTISAMPLE);
	}
}

void Grapher::initializeGL()
{
    this->makeCurrent();
	
#if defined(WIN32)
	std::cerr << "calling glewInit()" << std::endl;
	GLenum glewError = glewInit();
	if (glewError != GLEW_OK) {
		std::cerr << "GLEW error: " << glewGetErrorString(glewError) << std::endl;
	}
#endif
	
    glClearColor( 1.0, 1.0, 1.0, 1.0 );
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glEnable(GL_DEPTH_TEST);
	glEnable(GL_POINT_SMOOTH);
	glDisable(GL_MULTISAMPLE);
}

void Grapher::drawGraph(AbstractGraph* graph, int64_t offset, size_t* midpoint)
{
	fPoint high = maxValue_;
	fPoint low = -maxValue_;
	
	if (autoScale_) {
		low = minValue_ ;
	}
	
	fPoint leftMost = -(fPoint)((numValues_ - 1) / 2) + (fPoint)graphCenter_;
	fPoint rightMost = (fPoint)((numValues_ - 1) / 2) + (fPoint)graphCenter_;
	
	graph->draw(viewMatrix_, projectionMatrix_, width(), height(), high, low, leftMost, rightMost, rotated_, zoomFactor_);
}

void Grapher::drawLabels(QPainter& painter)
{
	// all labels are displayed at z = 0, so they are on top of the graph
	float zValue = 0.0f;
	
	fPoint high = maxValue_;
	fPoint low = -maxValue_;
	
	if (autoScale_) {
		low = minValue_ ;
	}
	
	painter.setPen(Qt::black);
	
/*	// when hovered, print the max value and the description of the hovered graph next to the y-axis
	if (hoveredGraph_) {
		char cText[1024];
		sprintf(cText, "max: %.4f %s", hoveredGraph_->getMaxValue(), hoveredGraph_->getDescription().c_str());
		drawText(zValue, TOP | HCENTER, std::string(cText), painter);
	}
*/	
	fPoint leftMost = -(fPoint)((numValues_ - 1) / 2) + (fPoint)graphCenter_;
	fPoint rightMost = (fPoint)((numValues_ - 1) / 2) + (fPoint)graphCenter_;
	
	int64_t valuesDisplayed = rightMost - leftMost;
	
	if (hovered_) {
		float modifier[2];
		modifier[0] = 0.01f * valuesDisplayed;
		modifier[1] = 0.01f * (high - low);
		
		// draw the value what we just hovered next to the cursor
		if (hoveredGraph_) {
			painter.setPen(Qt::blue);
			if (hoveredGraph_->getDataPtr() && (hoveredIndex_ < hoveredGraph_->getSize())) {
				char cText[1024];
				sprintf(cText, "%s : %.4f", hoveredGraph_->getDescription().c_str(), hoveredGraph_->getTop(hoveredIndex_).y() - hoveredGraph_->getBottom(hoveredIndex_).y());
				drawTextAt(zValue, hoveredPos_.x() + modifier[0], hoveredPos_.y() + modifier[1], std::string(cText), painter);
			}
		}
	}
	
	painter.setPen(Qt::black);
	
	if (spacing_ > 0) {
		// add the units to our unit graph, based on the levels that are actually displayed
		int level = (int)((float)log((float)(valuesDisplayed - 1) / 2.0f)) / log(spacing_ * 10.0f);
		
//		if (level > 0)
		{
			level = pow((spacing_ * 10.0f), (float)level); // if spacing is > 0 then the numbers are displayed at a factor of 10 more
			
			float startUnit = ((int)(leftMost / level) - 1) * level;
			float endUnit = startUnit + 2.0f * level;
			
			if (startUnit < leftMost && endUnit > rightMost) {
				level /= (spacing_ * 10.0f);
			}
			
			QMatrix4x4 pseudoView;
			pseudoView.lookAt(QVector3D((float)graphCenter_, 0.0, 0.0), QVector3D((float)graphCenter_, 0.0, -1.0), QVector3D(0.0, 1.0, 0.0));
			
			for (; startUnit < (int)rightMost; startUnit += level) {
				char cText[1024];
				sprintf(cText, "%.0f", startUnit);
				
				std::string text(cText);
				
				double x = graphCenter_;
				double y = 0.0;
				
				bool invertable;
				QMatrix4x4 viewprojectionInverse = (projectionMatrix_ * pseudoView).inverted(& invertable);
				
				// calculate the size of the text, so we can move it away from the Widget border properly
				QFontMetrics metric(painter.font());
				
				QString qText(text.c_str());
				
				float w = metric.width(qText);
				float h = metric.height();
				
				QRect size = metric.boundingRect(qText);
				
				double length;
				
				QVector4D mSize(w / width() * 2, h / height() * 2, 0.0, 0.0);
				QVector4D vSize = viewprojectionInverse * mSize;
				
				length = vSize.x();
				double horizontalBorder = (high - low) / height();
				
				if (rotated_) {
					y += low + length;// - (horizontalBorder + length);
				} else {
					y += low;
					y += horizontalBorder;
				}
				
				// transforming the text position into opengl coordinates, to properly place the text
				QVector4D coordinate((float)startUnit, y, zValue, 1.0f);
				QVector4D transformed = projectionMatrix_  * viewMatrix_ * coordinate;
				transformed /= transformed.w();
				transformed *= QVector4D(0.5f, 0.5f, 1.0f, 1.0f);
				transformed += QVector4D(0.5f, 0.5f, 0.0f, 0.0f);
				
				int ppx = transformed.x() * width();
				int ppy = height()* (1.0f - transformed.y()) - 5;
				
				painter.drawText(ppx, ppy, qText);
			}
		}
	}
}

void Grapher::drawText(fPoint m_Z, int m_Pos, std::string m_Text, QPainter& painter)
{
	// calculate the OpenGL position of the text, based on the text size in pixels, which is back transformed into OpenGL coordinates
	fPoint high = maxValue_;
	fPoint low = -maxValue_;
	
	if (autoScale_) {
		low = minValue_ ;
	}
	
	fPoint leftMost = -(fPoint)((numValues_ - 1) / 2);
	fPoint rightMost = (fPoint)((numValues_ - 1) / 2);
	
	double baseSize = 1.0;
    double horizontalBorder = (high - low)/ height();
	double verticalBorder = (rightMost - leftMost) / width();
	
	QMatrix4x4 pseudoView;
	pseudoView.lookAt(QVector3D((float)graphCenter_, 0.0, 0.0), QVector3D((float)graphCenter_, 0.0, -1.0), QVector3D(0.0, 1.0, 0.0));
	
	bool invertable;
	QMatrix4x4 viewprojectionInverse = (projectionMatrix_ * pseudoView).inverted(& invertable);
	
	QFontMetrics metric(painter.font());
	
	QString qText(m_Text.c_str());
	
	float w = metric.width(qText);
	float h = metric.height();
	
	QRect size = metric.boundingRect(qText);
	
	QVector4D mSize(w / width() * 2, h / height() * 2, 0.0, 0.0);
	QVector4D vSize = viewprojectionInverse * mSize;
	
	
	double fontSize = vSize.y();
	
	double length = vSize.x();
	
	// text size in OpenGL done
	
	double x = graphCenter_;
	double y = 0.0;
	
	// place the text along the widget border
	if (m_Pos & TOP) {
		if (rotated_) {
			y += high;// - (horizontalBorder);
		} else {
			y += high - (horizontalBorder + fontSize);
		}
	}
	if (m_Pos & BOTTOM) {
		if (rotated_) {
			y += low + length;// - (horizontalBorder + length);
		} else {
			y += low - (horizontalBorder);
		}
	}
	if (m_Pos & LEFT) {
		if (rotated_) {
			x -= -leftMost - (verticalBorder);
		} else {
			x -= -leftMost - verticalBorder;
		}
	}
	if (m_Pos & RIGHT) {
		if (rotated_) {
			x += rightMost - (verticalBorder + fontSize);
		} else {
			x += rightMost - (verticalBorder + length);
		}
	}
	
	// converting border position to the actual text start position
	QVector4D coordinate(x, y, m_Z, 1.0f);
	QVector4D transformed = projectionMatrix_  * viewMatrix_ * coordinate;
	transformed /= transformed.w();
	transformed *= QVector4D(0.5f, 0.5f, 1.0f, 1.0f);
	transformed += QVector4D(0.5f, 0.5f, 0.0f, 0.0f);
	
	int ppx = transformed.x() * width();
	int ppy = height()* (1.0f - transformed.y()) - 5;
	
	painter.drawText(ppx, ppy, qText);
}

void Grapher::drawTextAt(fPoint m_Z, fPoint m_X, fPoint m_Y, std::string m_Text, QPainter& painter)
{
	// calculate the OpenGL position of the text, based on the text size in pixels, which is back transformed into OpenGL coordinates
	fPoint high = maxValue_;
	fPoint low = -maxValue_;
	
	if (autoScale_) {
		low = minValue_ ;
	}
	
	fPoint leftMost = -(fPoint)((numValues_ - 1) / 2) + (fPoint)graphCenter_;
	fPoint rightMost = (fPoint)((numValues_ - 1) / 2) + (fPoint)graphCenter_;
	
	QMatrix4x4 pseudoView;
	pseudoView.lookAt(QVector3D((float)graphCenter_, 0.0, 0.0), QVector3D((float)graphCenter_, 0.0, -1.0), QVector3D(0.0, 1.0, 0.0));
	
	bool invertable;
	QMatrix4x4 viewprojectionInverse = (projectionMatrix_ * pseudoView).inverted(& invertable);
	
	QFontMetrics metric(painter.font());
	
	QString qText(m_Text.c_str());
	
	float w = metric.width(qText);
	float h = metric.height();
	
	QRect size = metric.boundingRect(qText);
	
	// text size in OpenGL done
	
	QVector4D mSize(w / width() * 2, h / height() * 2, 0.0, 0.0);
	if (rotated_) {
		mSize.setX(w / height() * 2);
		mSize.setY(h / width() * 2);
	}
	QVector4D vSize = viewprojectionInverse * mSize;
	
	if ((!rotated_ && (m_X + vSize.x()) > rightMost)) {
		m_X -= vSize.x();
	}
	
	if ((rotated_ && (m_X + vSize.y()) > rightMost)) {
		m_X -= vSize.y();
	}
	
	if ((!rotated_ && (m_Y + vSize.y()) > high)) {
		m_Y -= vSize.y();
	}
	
	if (rotated_ && (m_Y - vSize.x()) < low) {
		m_Y += vSize.x();
	}
	
	// convert the given position to the actual text start position
	QVector4D coordinate(m_X, m_Y, m_Z, 1.0f);
	QVector4D transformed = projectionMatrix_  * viewMatrix_ * coordinate;
	transformed /= transformed.w();
	transformed *= QVector4D(0.5f, 0.5f, 1.0f, 1.0f);
	transformed += QVector4D(0.5f, 0.5f, 0.0f, 0.0f);
	
	int ppx = transformed.x() * width();
	int ppy = height()* (1.0f - transformed.y()) - 5;
	
	painter.drawText(ppx, ppy, qText);
}

void Grapher::drawSelectedArea()
{
	if (!data_.empty()) {
		float zValue = (*(--data_.end()))->getOffset().z();
		fPoint high = maxValue_;
		fPoint low = -maxValue_;
		
		if(autoScale_) {
			low = minValue_ ;
		}
		
		glColor4f(0.7764f, 0.88627f, 1.0f, 1.0f);
		glBegin(GL_QUADS);
		glVertex3f(selectionRange_.first, low,  zValue);	// Bottom Left
		glVertex3f(selectionRange_.second, low,  zValue);	// Bottom Right
		glVertex3f(selectionRange_.second,  high,  zValue);	// Top Right
		glVertex3f(selectionRange_.first,  high,  zValue);	// Top Left
		glEnd();
	}
}

void Grapher::calculateHovered()
{
	bool invertable;
	QMatrix4x4 viewprojection = (projectionMatrix_ * viewMatrix_).inverted(& invertable);
	
	QVector3D tmp;
	if (rotated_) {
		tmp.setX(-vecPosition_.y());
		tmp.setY(vecPosition_.x());
	} else {
		tmp.setX(vecPosition_.x());
		tmp.setY(vecPosition_.y());
	}
	
	// transform hovered pos in [-1,1] to actual world pos
	QVector3D worldPos = viewprojection * tmp;
	
	// calculating approximate index of values, so we don't have to compare against the whole data arrays
	int64_t index = floor(worldPos.x() + 0.5);
	
	hoveredPos_.setX((float) index);
	hoveredTop_.setX((float) index);
	hoveredBottom_.setX((float) index);

//	hoveredPos_.setY(0.0f);
	hoveredPos_.setY(worldPos.y());
	hoveredTop_.setY(0.0f);
	hoveredBottom_.setY(0.0f);
	
	hoveredGraph_ = NULL;
	hovered_ = false;
	
	// find the data point closest to the world position
	for (std::set<AbstractGraph*, GraphComparator>::iterator iter = data_.begin(); iter != data_.end(); ++iter) {
		StackedGraph* stackedGraph = dynamic_cast<StackedGraph*>(*iter);
		if (stackedGraph && (stackedGraph->getSize() > (index - stackedGraph->getOffset().x()))) {
			int64_t realIndex = index - stackedGraph->getOffset().x();
			if (realIndex < 0) {
				continue;
			}

			if (stackedGraph->getBottom(realIndex).y() <= worldPos.y() && worldPos.y() <= stackedGraph->getTop(realIndex).y()) {
//				hoveredPos_.setY((stackedGraph->getBottom(realIndex).y() + stackedGraph->getTop(realIndex).y()) / 2.0);
				hoveredTop_.setY(stackedGraph->getTop(realIndex).y());
				hoveredBottom_.setY(stackedGraph->getBottom(realIndex).y());
				hoveredIndex_ = realIndex;
				hovered_ = true;
				hoveredGraph_= stackedGraph;
				break;
			}
		}
	}
}
