#ifndef Grapher_hpp
#define Grapher_hpp

#ifdef WIN32
typedef __int64 int64_t;
#endif

#include <QGLWidget>
#include <QVector3D>
#include <QVector4D>
#include <QMatrix4x4>
#include <QGLFramebufferObject>
#include <QTime>
#include <vector>
#include <set>
#include "AbstractGraph.hpp"
#include "Precision.hpp"

#include "ContinuousGraph.hpp"
#include "StackedGraph.hpp"
#include "UnitGraph.hpp"

class Grapher : public QGLWidget {
    Q_OBJECT
public:
	explicit Grapher(const std::string& xUnit, QWidget *parent = 0);
	
    virtual ~Grapher();
	
    virtual void resizeGL(int w, int h);
	
	// paint event override
    virtual void paintEvent(QPaintEvent* e);
	
	// function called when a mouse button was pressed
	virtual void mousePressEvent(QMouseEvent* e);
	
	// function called when a mouse button was released
	virtual void mouseReleaseEvent(QMouseEvent* e);
	
	// function called when a mouse button was double clicked
	virtual void mouseDoubleClickEvent(QMouseEvent* e);
	
	// function called when the mouse was moved
	virtual void mouseMoveEvent(QMouseEvent *e);
	
	// redraw the screen...this is one of the functions called by paintEvent
    virtual void redraw();
	
	// spacing: the frequency of the units, eg. unit every minValue + n * spacing
	// repeat: if the units have to be repeated
	UnitGraph* addUnitGraph(size_t numValues, fPoint spacing, bool repeat, QVector4D color = QVector4D(0.0, 0.0, 0.0, 1.0), float lineWidth = 1, QVector3D offset = QVector3D(0.0, 0.0, 0.0), const std::string& unit = std::string());

	// single color version
	template<class DataIterator>
	StackedGraph* addStackedGraph(DataIterator dataBegin, DataIterator dataEnd, DataIterator topBegin, QVector4D color = QVector4D(0.0, 0.0, 0.0, 1.0), float fillAlpha = 0.9, float lineWidth = 1, QVector3D offset = QVector3D(0.0, 0.0, 0.0), const std::string& unit = std::string())
	{
		StackedGraph* ndata = new StackedGraph(dataBegin, dataEnd, topBegin, color, fillAlpha, lineWidth, offset, unit);
		
		data_.insert(ndata);
		
		if (ndata->getMaxValue() > maxValue_) {
			maxValue_ = ndata->getMaxValue();
		}
		if (ndata->getMinValue() < minValue_) {
			minValue_ = ndata->getMinValue();
		}
		
		redrawPending_ = true;
		
		return ndata;
	}

	// varying color version
	template<class DataIterator, class ColorIterator>
	StackedGraph* addStackedGraph(DataIterator bottomBegin, DataIterator bottomEnd, DataIterator topBegin, ColorIterator colorsBegin, float fillAlpha = 0.9, float lineWidth = 1, QVector3D offset = QVector3D(0.0, 0.0, 0.0), const std::string& unit = std::string())
	{
		StackedGraph * ndata = new StackedGraph(bottomBegin, bottomEnd, topBegin, colorsBegin, fillAlpha, lineWidth, offset, unit);
		
		data_.insert(ndata);
		
		if (ndata->getMaxValue() > maxValue_) {
			maxValue_ = ndata->getMaxValue();
		}
		if (ndata->getMinValue() < minValue_) {
			minValue_ = ndata->getMinValue();
		}
		
		redrawPending_ = true;
		
		return ndata;
	}
	
	// to remove a graph
    void removeData(AbstractGraph* graph);
	
	// clears all the graphs from the grapher
    void clear();
	
	// function to respond to a mouse wheel event
    void wheelEvent(QWheelEvent* event);
	
	// saves the current view to an image
	void saveTo(QString filePath);
	
	// returns the font size of all the labels (in points)
	int getFontSize();
	
	// returns the number of values the grapher should display
	size_t getNumValues() const;
	
	// returns which value the grapher displays at its center
	size_t getCenter() const;
	
	// returns the range of the current selection
	std::pair<int, int> getSelection() const;
	
signals:
	void centerChanged(size_t newCenter);	// pan
	void numValuesChanged(size_t numValues);	// zoom
	void currentSelectedIndexes(std::pair<int, int> indexRange);	// signal start and end index of area where mouse was dragged (right mouse button pressed)

public slots:
    void center(size_t frame);	// center the given frame
    void setNumValues(size_t numValues);	// change the number of displayed values
	void setAntiAliasing(bool enable);
	void setRotated(bool rotated);
	void setFontSize(int pixel);	// this is now in points, NOT pixels!
	void setAutoScale(bool autoScale);	// sets if the grapher should scale depending on the set min/max value over all the graphs
	
	// sets if the display of the hovered value is only changed when the mouse is moved (off),
	// or if moving the center value of the grapher via the center slot also updates the hovered value (on)
	void setAutoHover(bool autoHover);
	
	void setSelectedArea(std::pair<float, float> area);
	void clearSelection();
	
	void setSelectionSnap(bool snap);	// enables snapping the selection to index values
	
private slots:
    void checkRedrawPending();	// slot receiving the timer event, redraws the graphs if necessary
	
private:
	// struct for defining a location along the window borders: it is used do put the text on the screen
	typedef enum
	{
		TOP				= 1,
		BOTTOM			= 2,
		LEFT			= 4,
		RIGHT			= 8,
		
		VCENTER			= TOP | BOTTOM,
		HCENTER			= LEFT | RIGHT,
		
		CENTER			= VCENTER | HCENTER
	}
	borderPos;
	
	// overridden paintGL function. triggered by redraw/paintEvent. does all the OpenGL calls
    virtual void paintGL();
	
    virtual void initializeGL();
	
	// midpoint: the current frame displayed at x = 0
    void drawGraph(AbstractGraph* graph, int64_t offset, size_t* midpoint);
	
	void drawLabels(QPainter& painter);
	
	// draws a given text at a certain position along the border
	void drawText(fPoint m_Z, int m_Pos, std::string m_Text, QPainter& painter);
	
	// draws a given text at a certain position
	void drawTextAt(fPoint m_Z, fPoint m_X, fPoint m_Y, std::string m_Text, QPainter& painter);
	
	// draws a GL_QUAD between the start and end position of selectionRange_
	void drawSelectedArea();
	
	// calculates the hovered point
	void calculateHovered();

	std::set<AbstractGraph*, GraphComparator> data_;	// vector of the graph data elements
	std::string xUnit_;	// the unit on the x axis
	
	size_t numValues_;	// the number of values displayed. it is always odd to have values exactly at x = 0
	size_t graphCenter_;	// the frame where the graph should have it's center
	float zoomFactor_;	// the zoom factor in vertical direction
	QVector3D openGLPos_;	// the position in xy space in opengl that has to be displayed
	
	QVector3D vecPosition_;	// the current mouse position while dragging
	QVector3D dragError_;	// error vector while dragging
	bool isFirstFrame_;	// if the current frame is the first frame in the movement
	bool pressed_;	// if the mouse button is pressed
	bool leftPressed_;	// if the left button is pressed
	
	QGLFramebufferObject* imageFBO_;	// the fbo to render the image into
	
	float valuedist_;	// the distance between values
	int windowOffset_;	// the number of values the cursor offsets from the centered value when moving
	
	bool hovered_;	// if we are hovering over a certain point
	QVector3D hoveredPos_;	// the position of the hovered point
	QVector3D hoveredTop_;	// the position of the hovered point
	QVector3D hoveredBottom_;	// the position of the hovered point
	
	size_t hoveredIndex_;	// the index of the current hovered pos
	StackedGraph* hoveredGraph_;	// the hovered graph
	
	bool antialiasing_;	// if the curves should be rendered using antialiasing
	
	size_t lastNumDivides_;	// how often the units of the last frame were divided
	QTime divideOccurredTimer_;	// timer started/reset when a divide occurres
	
	bool rotated_;	// if the display is rotated
	
	int fontSize_;	// the font size in points
	
	fPoint maxValue_;	// the maximum value of all data
	fPoint minValue_;	// the minimum value of all data
	
	bool autoScale_;	// if the grapher should automatically scale to the min/max value
	
	float near_;	// the near clipping plane
	float far_;	// the far clipping plane
	
	bool redrawPending_;	// if the graphs have to be redrawn
	QTimer* redrawTimer_;	// causes redrawPending_ to be checked
	
	bool autoHover_;	// if the hovered position is calculated each frame
	
	std::pair<float, float> selectionRange_;	// selected index range
	bool snapSelection_;	// if selection should snap to grid
	
	QMatrix4x4 viewMatrix_;
	QMatrix4x4 projectionMatrix_;
	
	float spacing_;	// the spacing of the units
};

#endif
