#include "StackedGraph.hpp"

#include <QGLWidget>
#include <cmath>
#include <iostream>
#include <QMatrix4x4>

void StackedGraph::draw(QMatrix4x4 & view, QMatrix4x4 & projection, size_t width, size_t height, fPoint maxValue, fPoint minValue, fPoint leftValue, fPoint rightValue, bool rotated, float zoomFactor)
{
	glPushMatrix();
	QMatrix4x4 modelView;
	glLineWidth(lineWidth_);
	
	// loop over all the tiles of our graph
	for (size_t i = 0; i < vertexBuffers_.size() ; ++i) {
		VertexBuffer * current = vertexBuffers_[i];
		
		// check if the current graph tile is actually visible
		float xmin = offset_.x() + current->getXMin();
		float xmax = offset_.x() + current->getXMax();
		if (xmax < leftValue || xmin > rightValue) {
			continue;
		}

		// translate the tile to its proper position
		modelView = view;
		modelView.translate((offset_ + current->getBufferOffset()));
		glLoadMatrixd(modelView.data());
		
		// draw the whole vertexbuffer
		current->bind();
		glDrawElements(GL_TRIANGLES, current->getNumberOfIndices(), GL_UNSIGNED_INT, (void*)0);
		current->release();
	}
	glPopMatrix();
}

void StackedGraph::generateIndicesAndRegenerateVertexBuffer()
{
	// build the vertex buffer(s) from a given data set:
	// depending on ContinousGraph::splitThreshold we might end up with several vertex buffers,
	// that are defined between x = [0, splitThreshold], and then moved into place
	std::vector<std::vector<int> > allIndices;
	std::vector<std::vector<QVector3D> > allPositions;
	std::vector<std::vector<QVector3D> > allTops;
	std::vector<std::vector<QVector4D> > allColors;
	std::vector<size_t> allMins;
	std::vector<size_t> allMaxs;
	std::vector<size_t> prevPos;
	std::vector<QVector3D> allBufferOffsets;
	
	std::vector<int> curIndices;
	std::vector<QVector3D> curPos;
	std::vector<QVector3D> curTop;
	std::vector<QVector4D> curColors;
	
	size_t reserveValues = std::min(positions_.size(), ContinuousGraph::splitThreshold);
	
	curIndices.reserve(reserveValues);
	curPos.reserve(reserveValues);
	curTop.reserve(reserveValues);
	curColors.reserve(reserveValues);
	
	int64_t counter = 0;
	size_t min = 0;
	size_t offsetIndex = 0;
	fPoint offsetValue = 0;


	// generating the positions. this time the actual (x,y) value and the corresponding color
	for (size_t i = 0; i < positions_.size() - 1; ++i) {
		curPos.push_back(positions_[i]);
		curPos.back().setX(curPos.back().x() - offsetValue);
		curTop.push_back(top_[i]);
		curTop.back().setX(curTop.back().x() - offsetValue);
		curColors.push_back(QVector4D(colors_[i].x(), colors_[i].y(), colors_[i].z(), fillAlpha_));
		
		counter++;
		
		// encountered a split
		if ((counter == ContinuousGraph::splitThreshold) && (i != (positions_.size() - 1))) {
			// add the starting values for the next split part, so we don't get holes in between vertexbuffers
			curPos.push_back(positions_[i + 1]);
			curPos.back().setX(curPos.back().x() - offsetValue);
			curTop.push_back(top_[i + 1]);
			curTop.back().setX(curTop.back().x() - offsetValue);
			curColors.push_back(QVector4D(colors_[i + 1].x(), colors_[i + 1].y(), colors_[i + 1].z(), fillAlpha_));
			
			// store the values of the now previous vertexbuffer
			allPositions.push_back(curPos);
			allTops.push_back(curTop);
			allColors.push_back(curColors);
			allBufferOffsets.push_back(QVector3D(offsetValue, 0.0f, 0.0f));
			++offsetIndex;
			offsetValue = offsetIndex * ContinuousGraph::splitThreshold;
			
			prevPos.push_back(curPos.size());
			
			// next values initialised
			curPos.clear();
			curTop.clear();
			curColors.clear();
			
			curPos.reserve(reserveValues);
			curTop.reserve(reserveValues);
			curColors.reserve(reserveValues);
			
			allMins.push_back(min);
			allMaxs.push_back(i);
			counter = 0;
			
			min = i;
		}
	}
	
	// store the values of the last vertexbuffer
	allPositions.push_back(curPos);
	allTops.push_back(curTop);
	allColors.push_back(curColors);
	allMins.push_back(min);
	allMaxs.push_back(positions_.size());
	allBufferOffsets.push_back(QVector3D(offsetValue, 0.0f, 0.0f));
	offsetIndex = 0;
	offsetValue = 0;
	
	prevPos.push_back(curPos.size());
	
	size_t index = 0;
	std::vector<std::vector<QVector3D> >::iterator singlePos = allPositions.begin() + index;
	std::vector<std::vector<QVector4D> >::iterator singleColor = allColors.begin() + index;
	
	counter = 0;
	
	// add the correspoding (x,0) value at the end of every array, same for color
	for (size_t i = 0; i < std::max(positions_.size() - 1, (size_t)0); ++i) {
		(*singlePos).push_back(QVector3D(top_[i].x() - offsetValue, top_[i].y(), top_[i].z()));
		(*singleColor).push_back(QVector4D(colors_[i].x(), colors_[i].y(), colors_[i].z(), fillAlpha_));
		
		counter++;
		
		// encountered a split
		if (counter == ContinuousGraph::splitThreshold) {
			// add first value of next vertexbuffer at end, so we are complete
			(*singlePos).push_back(QVector3D(top_[i + 1].x() - offsetValue, top_[i + 1].y(), top_[i + 1].z()));
			(*singleColor).push_back(QVector4D(colors_[i + 1].x(), colors_[i + 1].y(), colors_[i + 1].z(), fillAlpha_));

			++offsetIndex;
			offsetValue = offsetIndex * ContinuousGraph::splitThreshold;
			
			// switch to the data for the next vertexbuffer
			index++;
			singlePos = allPositions.begin() + index;
			singleColor = allColors.begin() + index;
			
			counter = 0;
		}
	}
	
		
	allIndices.resize(allPositions.size());
	
	// generate the indices
	for (size_t j = 0; j < allPositions.size(); ++j) {
		// looping over all values in positions
		size_t prevPosSize = prevPos[j];
		for (size_t i = 0; i < std::max(prevPosSize - 1, (size_t)0); ++i) {
			// build the indexbuffer for the two triangles forming the filled region under the curve
			allIndices[j].push_back(i + prevPosSize);
			allIndices[j].push_back(i + prevPosSize + 1);
			allIndices[j].push_back(i);
			
			allIndices[j].push_back(i);
			allIndices[j].push_back(i + prevPosSize + 1);
			allIndices[j].push_back(i + 1);
		}
		
		// done with the indices, build the actual vertexbuffer
		VertexBuffer * current = new VertexBuffer(allIndices[j], allPositions[j], allColors[j]);
		current->setXMin(allMins[j]);
		current->setXMax(allMaxs[j]);
		current->setBufferOffset(allBufferOffsets[j]);
		
		vertexBuffers_.push_back(current);
	}
}

QVector3D StackedGraph::getBottom(size_t i) const
{
	return positions_[i];
}

QVector3D StackedGraph::getTop(size_t i) const
{
	return top_[i];
}

void StackedGraph::calculateExtremeValues()
{
	maxValue_ = std::numeric_limits<fPoint>::min();
	minValue_ = std::numeric_limits<fPoint>::max();

	for (size_t i = 0; i < positions_.size(); ++i) {
		if (top_[i].y() > maxValue_) {
			maxValue_ = top_[i].y();
		}
		if (positions_[i].y() < minValue_) {
			minValue_ = positions_[i].y();
		}
	}
}
